//==========================================================================
//
//      io/disk/disk.c
//
//      High level disk driver
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
// Copyright (C) 2003 Savin Zlobec 
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):     savin 
// Date:          2003-06-10
// Purpose:       Top level disk driver
// Description: 
//
//####DESCRIPTIONEND####
//
//==========================================================================

#include <pkgconf/io.h>
#include <pkgconf/io_disk.h>

#include <cyg/io/io.h>
#include <cyg/io/devtab.h>
#include <cyg/io/disk.h>
#include <cyg/infra/cyg_ass.h>      // assertion support
#include <cyg/infra/diag.h>         // diagnostic output

// ---------------------------------------------------------------------------

#define DEBUG 1

// ---------------------------------------------------------------------------

static Cyg_ErrNo disk_bwrite(cyg_io_handle_t handle, const void *buf, 
                             cyg_uint32 *len, cyg_uint32 pos);
static Cyg_ErrNo disk_bread(cyg_io_handle_t handle, void *buf, 
                            cyg_uint32 *len, cyg_uint32 pos);
static Cyg_ErrNo disk_select(cyg_io_handle_t handle, cyg_uint32 which, 
                             CYG_ADDRWORD info);
static Cyg_ErrNo disk_get_config(cyg_io_handle_t handle, cyg_uint32 key, 
                                 void *buf, cyg_uint32 *len);
static Cyg_ErrNo disk_set_config(cyg_io_handle_t handle, cyg_uint32 key, 
                                 const void *buf, cyg_uint32 *len);

BLOCK_DEVIO_TABLE(cyg_io_disk_devio,
                  disk_bwrite,
                  disk_bread,
                  disk_select,
                  disk_get_config,
                  disk_set_config
);

static bool disk_init(disk_channel *chan, cyg_disk_geom_t *geom);

DISK_CALLBACKS(cyg_io_disk_callbacks, 
               disk_init
); 

// ---------------------------------------------------------------------------

static bool 
disk_init(disk_channel *chan, cyg_disk_geom_t *geom)
{
    if (!chan->init)
    {
        chan->init = true;
        chan->config.block_size = 512;
        chan->config.blocks_num = geom->sectors_num;

#ifdef DEBUG
        diag_printf("disk init block_size=%d blocks_num=%d\n", 
                    chan->config.block_size, chan->config.blocks_num);
#endif
 
    }
    return true;  
}

// ---------------------------------------------------------------------------

static Cyg_ErrNo 
disk_bwrite(cyg_io_handle_t  handle, 
            const void      *buf, 
            cyg_uint32      *len,
            cyg_uint32       pos)
{
    cyg_devtab_entry_t *t         = (cyg_devtab_entry_t *) handle;
    disk_channel       *chan      = (disk_channel *) t->priv;
    disk_funs          *funs      = chan->funs;
    cyg_disk_info_t    *disk_info = &chan->config;
    cyg_int32  size  = *len;
    cyg_int32  block = pos;
    cyg_uint8 *bbuf  = (cyg_uint8 * const) buf;
    Cyg_ErrNo  res   = ENOERR;

#ifdef DEBUG
    diag_printf("disk write block=%d len=%d buf=%p\n", pos, *len, buf); 
#endif
    
    while (size > 0)
    {
        (funs->write)(chan, block, (void*)bbuf);
        if (ENOERR != res)
            break;
 
        bbuf += disk_info->block_size;
        block++;
        size--;
    }

    *len -= size;
    return res;
}

// ---------------------------------------------------------------------------

static Cyg_ErrNo 
disk_bread(cyg_io_handle_t  handle, 
           void            *buf, 
           cyg_uint32      *len,
           cyg_uint32       pos)
{
    cyg_devtab_entry_t *t         = (cyg_devtab_entry_t *) handle;
    disk_channel       *chan      = (disk_channel *) t->priv;
    disk_funs          *funs      = chan->funs;
    cyg_disk_info_t    *disk_info = &chan->config;
    cyg_int32  size  = *len;
    cyg_int32  block = pos;
    cyg_uint8 *bbuf  = buf;
    Cyg_ErrNo  res   = ENOERR;
 
#ifdef DEBUG
    diag_printf("disk read block=%d len=%d buf=%p\n", pos, *len, buf); 
#endif
 
    while (size > 0)
    {
        res = (funs->read)(chan, block, (void*)bbuf);
        if (ENOERR != res)
            break;
        
        bbuf += disk_info->block_size;
        block++;
        size--;
    }
    
    *len -= size;
    return res;
}

// ---------------------------------------------------------------------------

static cyg_bool
disk_select(cyg_io_handle_t handle, cyg_uint32 which, CYG_ADDRWORD info)
{
    return true;
}

// ---------------------------------------------------------------------------

static Cyg_ErrNo 
disk_get_config(cyg_io_handle_t  handle, 
                cyg_uint32       key, 
                void            *xbuf,
                cyg_uint32      *len)
{
    cyg_devtab_entry_t *t    = (cyg_devtab_entry_t *) handle;
    disk_channel       *chan = (disk_channel *) t->priv;
    cyg_disk_info_t    *buf  = (cyg_disk_info_t *) xbuf;
    disk_funs          *funs = chan->funs;
    Cyg_ErrNo res = ENOERR;
 
#ifdef DEBUG
    diag_printf("disk get config\n"); 
#endif
    
    switch (key) {
    case CYG_IO_GET_CONFIG_DISK_INFO:
        if (*len < sizeof(cyg_disk_info_t)) {
            return -EINVAL;
        }
        *buf = chan->config;
        *len = sizeof(chan->config);
        break;       

    default:
        // pass down to lower layers
        res = (funs->get_config)(chan, key, xbuf, len);
    }
   
    return res;
}

// ---------------------------------------------------------------------------

static Cyg_ErrNo 
disk_set_config(cyg_io_handle_t  handle, 
                cyg_uint32       key, 
                const void      *xbuf, 
                cyg_uint32      *len)
{
    cyg_devtab_entry_t *t    = (cyg_devtab_entry_t *) handle;
    disk_channel       *chan = (disk_channel *) t->priv;
    disk_funs          *funs = chan->funs;

#ifdef DEBUG
    diag_printf("disk set config\n"); 
#endif
 
    // pass down to lower layers
    return (funs->set_config)(chan, key, xbuf, len);
}

// ---------------------------------------------------------------------------
// EOF disk.c
