//==========================================================================
//
//      io/disk/disk_bio.c
//
//      Disk byte IO driver
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 2003 Savin Zlobec 
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):     savin 
// Date:          2003-08-19
// Description: 
//
//####DESCRIPTIONEND####
//
//==========================================================================

#include <pkgconf/io.h>
#include <pkgconf/io_disk.h>

#include <cyg/io/io.h>
#include <cyg/io/devtab.h>
#include <cyg/io/disk.h>
#include <cyg/infra/cyg_ass.h>      // assertion support
#include <cyg/infra/diag.h>         // diagnostic output
#include <cyg/kernel/kapi.h>

#include <stdlib.h> // malloc

// --------------------------------------------------------------------

#define DEBUG 1

// --------------------------------------------------------------------

typedef struct {
    cyg_bool         init;            // true if initialized
    cyg_uint32       current_block;   // current block 
    cyg_uint32       block_size;      // block size
    cyg_uint32       block_size_log2; // block size log2
    cyg_uint8       *block_buf;       // block buffer
    cyg_io_handle_t  dev_h;           // disk device handle
} bio_info_t;

// --------------------------------------------------------------------

static bool bio_init(struct cyg_devtab_entry *tab);

static Cyg_ErrNo bio_lookup(struct cyg_devtab_entry **tab, 
                            struct cyg_devtab_entry  *sub_tab,
                            const char               *name);

static Cyg_ErrNo bio_bwrite(cyg_io_handle_t  handle, 
                            const void      *buf,
                            cyg_uint32      *len, 
                            cyg_uint32       pos);

static Cyg_ErrNo bio_bread(cyg_io_handle_t  handle, 
                           void            *buf,
                           cyg_uint32      *len, 
                           cyg_uint32       pos);

static Cyg_ErrNo bio_select(cyg_io_handle_t handle, 
                            cyg_uint32      which,
                            CYG_ADDRWORD    info);

static Cyg_ErrNo bio_get_config(cyg_io_handle_t  handle, 
                                cyg_uint32       key,
                                void            *xbuf, 
                                cyg_uint32      *len);

static Cyg_ErrNo bio_set_config(cyg_io_handle_t  handle, 
                                cyg_uint32       key,
                                const void      *xbuf, 
                                cyg_uint32      *len);

BLOCK_DEVIO_TABLE(bio_devio,
                  bio_bwrite,
                  bio_bread,
                  bio_select,
                  bio_get_config,
                  bio_set_config
);

#define DISK_BIO_INSTANCE(_number_)                            \
static bio_info_t bio_info##_number_ = {                       \
    init:          false,                                      \
    current_block: 0,                                          \
    block_buf:     NULL                                        \
};                                                             \
DEVTAB_ENTRY(bio##_number_,                                    \
             CYGDAT_IO_DISK_BIO_DISK##_number_##_NAME,         \
             CYGDAT_IO_DISK_BIO_DISK##_number_##_DEV,          \
             &bio_devio,                                       \
             bio_init,                                         \
             bio_lookup,                                       \
             &bio_info##_number_);


#ifdef CYGPKG_IO_DISK_BIO_DISK0
DISK_BIO_INSTANCE(0);
#endif

// --------------------------------------------------------------------

static cyg_uint32
get_val_log2(cyg_uint32 val)
{
    cyg_uint32 i, log2;

    i = val;
    log2 = 0;
    while (0 == (i & 1))
    {
        i >>= 1;
        log2++;
    }
    if (i != 1)
        return 0;
    else
        return log2;
}

// --------------------------------------------------------------------

static bool 
bio_init(struct cyg_devtab_entry *tab)
{
    return true;
}

static Cyg_ErrNo 
bio_lookup(struct cyg_devtab_entry **tab, 
           struct cyg_devtab_entry  *sub_tab,
           const char               *name)
{
    cyg_io_handle_t  chan = (cyg_io_handle_t)sub_tab;
    bio_info_t      *bio  = (bio_info_t *)(*tab)->priv;

    if (!bio->init)
    {
        cyg_disk_info_t  disk_info;
        int len = sizeof(cyg_disk_info_t);

        bio->init = true;
 
        cyg_io_get_config(chan, CYG_IO_GET_CONFIG_DISK_INFO, 
                          (void *)&disk_info, &len);
       
        bio->block_buf = (cyg_uint8 *)malloc(disk_info.block_size);
        if (NULL == bio->block_buf)
            return -ENOMEM;
       
        bio->dev_h           = chan;
        bio->block_size      = disk_info.block_size;
        bio->block_size_log2 = get_val_log2(disk_info.block_size);

#ifdef DEBUG
        diag_printf("disk bio init block_size=%d (log2=%d)\n", 
                    bio->block_size, bio->block_size_log2); 
#endif
 
        if (0 == bio->block_size_log2)
            return -EINVAL;
    }
    
    return ENOERR;
}

static Cyg_ErrNo
bio_bread(cyg_io_handle_t  handle,
          void            *buf,
          cyg_uint32      *len,
          cyg_uint32       pos)
{
    cyg_devtab_entry_t *t   = (cyg_devtab_entry_t *) handle;
    bio_info_t         *bio = (bio_info_t *) t->priv;
    cyg_int32 size, bpos, bnum;
    Cyg_ErrNo res = ENOERR;

    size = *len;
    bpos = pos;
    bnum = bio->current_block;

    if (bpos >= bio->block_size)
    {
        bnum += bpos >> bio->block_size_log2;
        bpos  = bpos & (bio->block_size - 1); 
    }
    
#ifdef DEBUG
    diag_printf("disk bio read len=%d bpos=%d bnum=%d (%d)\n", 
                *len, bpos, bnum, bio->current_block); 
#endif

    while (size > 0)
    {
        int csize, l;
        
        if ((size + bpos) > bio->block_size)
            csize = bio->block_size - bpos;
        else
            csize = size;
        
        l = 1;
        res = cyg_io_bread(bio->dev_h, (void*)bio->block_buf, &l, bnum);
        if (ENOERR != res)
            break;
        memcpy(buf, (void*)&(bio->block_buf[bpos]), csize);
       
        buf   = &(((cyg_uint8 *)buf)[csize]);
        size -= csize;
        bpos  = 0;
        bnum++;
    }

    *len -= size;
    return res;
}

static Cyg_ErrNo
bio_bwrite(cyg_io_handle_t  handle,
           const void      *buf,
           cyg_uint32      *len,
           cyg_uint32       pos)
{
    cyg_devtab_entry_t *t   = (cyg_devtab_entry_t *) handle;
    bio_info_t         *bio = (bio_info_t *) t->priv;
    cyg_int32 size, bpos, bnum;
    Cyg_ErrNo res = ENOERR;

    size = *len;
    bpos = pos;
    bnum = bio->current_block;

    if (bpos >= bio->block_size)
    {
        bnum += bpos >> bio->block_size_log2;
        bpos  = bpos & (bio->block_size - 1); 
    }
    
#ifdef DEBUG
    diag_printf("disk bio write len=%d bpos=%d bnum=%d (%d)\n", 
                *len, bpos, bnum, bio->current_block); 
#endif

    while (size > 0)
    {
        int csize, l;
        
        if ((size + bpos) > bio->block_size)
            csize = bio->block_size - bpos;
        else
            csize = size;
        
        if (bio->block_size != csize || bpos != 0)
        {
            l = 1;
            res = cyg_io_bread(bio->dev_h, (void*)bio->block_buf, &l, bnum);
            if (ENOERR != res)
                break;
        }
        memcpy((void*)&(bio->block_buf[bpos]), buf, csize);
        l = 1;
        res = cyg_io_bwrite(bio->dev_h, (void*)bio->block_buf, &l, bnum);
        if (ENOERR != res)
           break;
        
        buf   = &(((cyg_uint8 *)buf)[csize]);
        size -= csize;
        bpos  = 0;
        bnum++;
    }

    *len -= size;
    return res;
}

static cyg_bool
bio_select(cyg_io_handle_t handle, cyg_uint32 which, CYG_ADDRWORD info)
{
    cyg_devtab_entry_t *t   = (cyg_devtab_entry_t *) handle;
    bio_info_t         *bio = (bio_info_t *) t->priv;

    // Just pass it on to next driver level
    return cyg_io_select(bio->dev_h, which, info);
}

static Cyg_ErrNo 
bio_get_config(cyg_io_handle_t  handle, 
               cyg_uint32       key, 
               void            *xbuf, 
               cyg_uint32      *len)
{
    cyg_devtab_entry_t *t   = (cyg_devtab_entry_t *) handle;
    bio_info_t         *bio = (bio_info_t *) t->priv;
    cyg_uint32         *buf = (cyg_uint32 *) xbuf;
    Cyg_ErrNo res = ENOERR;

#ifdef DEBUG
    diag_printf("disk bio get config\n"); 
#endif

    switch (key) {
    case CYG_IO_GET_CONFIG_DISK_BIO_CURRENT_BLOCK:
        if (*len < sizeof(cyg_uint32)) {
            return -EINVAL;
        }
        *buf = bio->current_block;
        *len = sizeof(cyg_uint32);
        break;

    default:
        // Pass it on to next driver level
        res = cyg_io_get_config(bio->dev_h, key, xbuf, len);
    }

    return res;
}

static Cyg_ErrNo 
bio_set_config(cyg_io_handle_t  handle, 
               cyg_uint32       key, 
               const void      *xbuf, 
               cyg_uint32      *len)
{
    cyg_devtab_entry_t *t   = (cyg_devtab_entry_t *) handle;
    bio_info_t         *bio = (bio_info_t *) t->priv;
    cyg_uint32         *buf = (cyg_uint32 *) xbuf;
    Cyg_ErrNo res = ENOERR;

#ifdef DEBUG
    diag_printf("disk bio set config\n"); 
#endif

    switch (key) {
    case CYG_IO_SET_CONFIG_DISK_BIO_CURRENT_BLOCK:
        if (*len != sizeof(cyg_uint32)) {
            return -EINVAL;
        }
        bio->current_block = *buf;
        break;

    default:
        // Pass it on to next driver level
        res = cyg_io_set_config(bio->dev_h, key, xbuf, len);
    }

    return res;
}

// --------------------------------------------------------------------
// EOF disk_bio.c
