//==========================================================================
//
//      io/disk/disk_cache.c
//
//      Cached disk I/O driver
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 2003 Savin Zlobec 
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):     savin 
// Date:          2003-08-18
// Description: 
//
//####DESCRIPTIONEND####
//
//==========================================================================

#include <pkgconf/io.h>
#include <pkgconf/io_disk.h>

#include <cyg/io/io.h>
#include <cyg/io/devtab.h>
#include <cyg/io/disk.h>
#include <cyg/infra/cyg_ass.h>      // assertion support
#include <cyg/infra/diag.h>         // diagnostic output
#include <cyg/kernel/kapi.h>

#include <linux/rbtree.h>
#include <linux/list.h>

// --------------------------------------------------------------------

#define DEBUG 1

#define STAT_INIT(_dc_) do {    \
        dc->stat.n_gets   = 0;  \
        dc->stat.n_reads  = 0;  \
        dc->stat.n_writes = 0;  \
    } while (0)    

#define STAT(_dc_, _group_) \
    ((_dc_)->stat._group_++)

// --------------------------------------------------------------------
  
typedef struct {
    cyg_bool              init;       // true if cache is initialized
    cyg_io_handle_t       dev_h;      // disk device handle
    struct list_head      list_head;  // head of block list 
    struct rb_root        rb_root;    // block red-black tree root
    cyg_uint32            block_size; // block size
    cyg_uint8            *mem_base;   // memory base
    cyg_uint32            mem_size;   // memory size
    cyg_mempool_fix       mem_pool;   // memory pool struct
    cyg_handle_t          mem_pool_h; // memory pool handle
    cyg_disk_cache_stat_t stat;       // statistics
} dcache_info_t;

typedef struct {
    struct list_head  list_node; // list node
    struct rb_node    rb_node;   // red-black tree node
    cyg_uint32        num;       // block number
    cyg_bool          modified;  // is this block data modified (needs write)
    unsigned char     data[0];   // block data
} dcache_block_t;

// --------------------------------------------------------------------

static bool dcache_init(struct cyg_devtab_entry *tab);

static Cyg_ErrNo dcache_lookup(struct cyg_devtab_entry **tab, 
                               struct cyg_devtab_entry  *sub_tab,
                               const char               *name);

static Cyg_ErrNo dcache_bwrite(cyg_io_handle_t  handle, 
                               const void      *buf,
                               cyg_uint32      *len, 
                               cyg_uint32       pos);

static Cyg_ErrNo dcache_bread(cyg_io_handle_t  handle, 
                              void            *buf,
                              cyg_uint32      *len, 
                              cyg_uint32       pos);

static Cyg_ErrNo dcache_select(cyg_io_handle_t handle, 
                               cyg_uint32      which,
                               CYG_ADDRWORD    info);

static Cyg_ErrNo dcache_get_config(cyg_io_handle_t  handle, 
                                   cyg_uint32       key,
                                   void            *xbuf, 
                                   cyg_uint32      *len);

static Cyg_ErrNo dcache_set_config(cyg_io_handle_t  handle, 
                                   cyg_uint32       key,
                                   const void      *xbuf, 
                                   cyg_uint32      *len);

BLOCK_DEVIO_TABLE(dcache_devio,
                  dcache_bwrite,
                  dcache_bread,
                  dcache_select,
                  dcache_get_config,
                  dcache_set_config
);

#define DISK_CACHE_INSTANCE(_number_)                           \
static cyg_uint8 dcache_mem_base##_number_##[CYGDAT_IO_DISK_CACHE_DISK##_number_##_MEMSIZE]; \
static dcache_info_t dcache_info##_number_ = {                  \
    init:       false,                                          \
    mem_base:   dcache_mem_base##_number_,                      \
    mem_size:   CYGDAT_IO_DISK_CACHE_DISK##_number_##_MEMSIZE   \
};                                                              \
DEVTAB_ENTRY(dcache##_number_,                                  \
             CYGDAT_IO_DISK_CACHE_DISK##_number_##_NAME,        \
             CYGDAT_IO_DISK_CACHE_DISK##_number_##_DEV,         \
             &dcache_devio,                                     \
             dcache_init,                                       \
             dcache_lookup,                                     \
             &dcache_info##_number_);


#ifdef CYGPKG_IO_DISK_CACHE_DISK0
DISK_CACHE_INSTANCE(0);
#endif

// --------------------------------------------------------------------

static dcache_block_t *
rb_find_block(dcache_info_t *dc, cyg_uint32 num)
{
    struct rb_node *node  = dc->rb_root.rb_node;
    dcache_block_t *block = NULL;
    
    while (NULL != node)
    {
        block = rb_entry(node, dcache_block_t, rb_node);

        if (block->num == num)
            return block;

        node = (block->num > num) ? node->rb_left : node->rb_right;
    }
    return NULL;
}

static void
rb_add_block(dcache_info_t *dc, dcache_block_t *block)
{
    struct rb_node *node = dc->rb_root.rb_node;    

    if (NULL == node)
    {
        rb_link_node(&block->rb_node, NULL, &dc->rb_root.rb_node);
    }
    else
    {
        struct rb_node **link;
        dcache_block_t  *b = NULL;
        
        while (NULL != node)
        {
            b = rb_entry(node, dcache_block_t, rb_node);

            CYG_ASSERTC(b->num != block->num);

            link = (b->num > block->num) ? &node->rb_left : &node->rb_right;
            node = *link;
        }
        rb_link_node(&block->rb_node, &b->rb_node, link);
    }
    rb_insert_color(&block->rb_node, &dc->rb_root);
}

static __inline__ void
rb_del_block(dcache_info_t *dc, dcache_block_t *block)
{
    rb_erase(&block->rb_node, &dc->rb_root);
}

// --------------------------------------------------------------------

static __inline__ void
list_add_block(dcache_info_t *dc, dcache_block_t *block)
{
    list_add(&block->list_node, &dc->list_head);
}

static __inline__ void
list_del_block(dcache_info_t *dc, dcache_block_t *block)
{
    list_del(&block->list_node);
}

static __inline__ dcache_block_t *
list_get_last_block(dcache_info_t *dc)
{
    return(list_entry(dc->list_head.prev, dcache_block_t, list_node));
}

static void
list_move_block_to_head(dcache_info_t *dc, dcache_block_t *block)
{
    list_del(&block->list_node);
    list_add(&block->list_node, &dc->list_head);
}

// --------------------------------------------------------------------

static Cyg_ErrNo
dcache_sync_block(dcache_info_t *dc, dcache_block_t *block)
{
    Cyg_ErrNo ret = ENOERR;

    if (block->modified)
    {
        int len = 1;

#ifdef DEBUG
        diag_printf("disk cache writting block=%d to disk\n", block->num); 
#endif
 
        STAT(dc, n_writes);

        ret = cyg_io_bwrite(dc->dev_h, (void *)block->data, &len, block->num);

        if (ENOERR == ret)
            block->modified = false;
    }
    
    return ret;
}

static Cyg_ErrNo 
dcache_sync(dcache_info_t *dc)
{
    struct list_head *node = dc->list_head.next;
    dcache_block_t   *block;
    Cyg_ErrNo ret = ENOERR;
 
#ifdef DEBUG
    diag_printf("disk cache sync\n"); 
#endif
    
    while (node != &dc->list_head)
    {
        block = list_entry(node, dcache_block_t, list_node);

        ret = dcache_sync_block(dc, block);
        if (ENOERR != ret)
           return ret;
        
        node = node->next;
    }

    return ret;
}

static Cyg_ErrNo 
dcache_get_block(dcache_info_t *dc, cyg_uint32 num, dcache_block_t **dblock)
{
    dcache_block_t* block;
    Cyg_ErrNo ret = ENOERR;
    int len;
 
#ifdef DEBUG
    diag_printf("disk cache get block=%d\n", num); 
#endif
    
    STAT(dc, n_gets);

    block = rb_find_block(dc, num);
    
    if (block != NULL)
    {
#ifdef DEBUG
    diag_printf("disk cache block=%d found in cache (dblock=%p)\n", 
                num, block); 
#endif
        list_move_block_to_head(dc, block);
        *dblock = block;
        return ret;
    }

#ifdef DEBUG
    diag_printf("disk cache block=%d NOT found in cache\n", num); 
#endif

    block = (dcache_block_t *)cyg_mempool_fix_try_alloc(dc->mem_pool_h);

    if (NULL == block)
    {
        CYG_ASSERTC(!list_empty(&dc->list_head));
        
        block = list_get_last_block(dc);

#ifdef DEBUG
        diag_printf("disk cache reusing block=%d space\n", block->num); 
#endif

        ret = dcache_sync_block(dc, block);
        if (ENOERR != ret)
            return ret;
        
        list_del_block(dc, block);
        rb_del_block(dc, block);
    }

    block->num      = num;
    block->modified = false;

#ifdef DEBUG
    diag_printf("disk cache reading block=%d from disk\n", block->num); 
#endif

    STAT(dc, n_reads);

    len = 1;
    ret = cyg_io_bread(dc->dev_h, (void *)block->data, &len, block->num);
    if (ENOERR != ret)
    {
        cyg_mempool_fix_free(dc->mem_pool_h, block);
        return ret;
    }
    
    rb_add_block(dc, block);
    list_add_block(dc, block);
    
    *dblock = block;
    return ret;
}

// --------------------------------------------------------------------

static bool 
dcache_init(struct cyg_devtab_entry *tab)
{
    return true;
}

static Cyg_ErrNo 
dcache_lookup(struct cyg_devtab_entry **tab, 
              struct cyg_devtab_entry  *sub_tab,
              const char               *name)
{
    cyg_io_handle_t  chan = (cyg_io_handle_t)sub_tab;
    dcache_info_t   *dc   = (dcache_info_t *)(*tab)->priv;
    cyg_uint32 block_size;
    
    if (!dc->init)
    {
        cyg_disk_info_t disk_info;
        int len = sizeof(cyg_disk_info_t);
        
        cyg_io_get_config(chan, CYG_IO_GET_CONFIG_DISK_INFO, 
                          (void *)&disk_info, &len);
        
        dc->block_size = disk_info.block_size;

#ifdef DEBUG
        diag_printf("disk cache init mem_size=%d block_size=%d\n",
                    dc->mem_size, dc->block_size); 
#endif

        dc->init  = true;
        dc->dev_h = chan;
        
        dc->rb_root = RB_ROOT;
        INIT_LIST_HEAD(&dc->list_head);

        block_size = dc->block_size + sizeof(dcache_block_t);

        cyg_mempool_fix_create(dc->mem_base, dc->mem_size, block_size,
                               &dc->mem_pool_h, &dc->mem_pool);

        STAT_INIT(dc);
    }
    
    return ENOERR;
}

static Cyg_ErrNo
dcache_bread(cyg_io_handle_t  handle,
             void            *buf,
             cyg_uint32      *len,
             cyg_uint32       pos)
{
    cyg_devtab_entry_t *t  = (cyg_devtab_entry_t *) handle;
    dcache_info_t      *dc = (dcache_info_t *) t->priv;
    dcache_block_t     *dblock;
    cyg_int32  size  = *len;
    cyg_int32  block = pos;
    cyg_uint8 *bbuf  = buf;
    Cyg_ErrNo  res   = ENOERR;

#ifdef DEBUG
    diag_printf("disk cache read block=%d len=%d buf=%p\n", pos, *len, buf); 
#endif

    while (size > 0)
    {
        res = dcache_get_block(dc, block, &dblock); 
        if (ENOERR != res)
            break;
        
        memcpy((void *)bbuf, (void *)dblock->data, dc->block_size);

        bbuf += dc->block_size;
        block++;
        size--;
    }

    *len -= size;
    return res;
}

static Cyg_ErrNo
dcache_bwrite(cyg_io_handle_t  handle,
              const void      *buf,
              cyg_uint32      *len,
              cyg_uint32       pos)
{
    cyg_devtab_entry_t *t  = (cyg_devtab_entry_t *) handle;
    dcache_info_t      *dc = (dcache_info_t *) t->priv;
    dcache_block_t     *dblock;
    cyg_int32  size  = *len;
    cyg_int32  block = pos;
    cyg_uint8 *bbuf  = (cyg_uint8 * const) buf;
    Cyg_ErrNo  res   = ENOERR;

#ifdef DEBUG
    diag_printf("disk cache write block=%d len=%d buf=%p\n", pos, *len, buf); 
#endif

    while (size > 0)
    {
        res = dcache_get_block(dc, block, &dblock); 
        if (ENOERR != res)
            break;
        
        memcpy((void *)dblock->data, (void *)bbuf, dc->block_size);
        dblock->modified = true;

        bbuf += dc->block_size;
        block++;
        size--;
    }

    *len -= size;
    return res;
}

static cyg_bool
dcache_select(cyg_io_handle_t handle, cyg_uint32 which, CYG_ADDRWORD info)
{
    cyg_devtab_entry_t *t  = (cyg_devtab_entry_t *) handle;
    dcache_info_t      *dc = (dcache_info_t *) t->priv;

    // Just pass it on to next driver level
    return cyg_io_select(dc->dev_h, which, info);
}

static Cyg_ErrNo 
dcache_get_config(cyg_io_handle_t  handle, 
                  cyg_uint32       key, 
                  void            *xbuf, 
                  cyg_uint32      *len)
{
    cyg_devtab_entry_t    *t   = (cyg_devtab_entry_t *) handle;
    dcache_info_t         *dc  = (dcache_info_t *) t->priv;
    cyg_disk_cache_stat_t *buf = (cyg_disk_cache_stat_t *) xbuf;
    Cyg_ErrNo res = ENOERR;

#ifdef DEBUG
    diag_printf("disk cache get config\n"); 
#endif

    switch (key) {
    case CYG_IO_GET_CONFIG_DISK_CACHE_STAT:
        if (*len < sizeof(cyg_disk_cache_stat_t)) {
            return -EINVAL;
        }
        *buf = dc->stat;
        *len = sizeof(dc->stat);
        break;

    default:
        // Pass it on to next driver level
        res = cyg_io_get_config(dc->dev_h, key, xbuf, len);
    }

    return res;
}

static Cyg_ErrNo 
dcache_set_config(cyg_io_handle_t  handle, 
                  cyg_uint32       key, 
                  const void      *xbuf, 
                  cyg_uint32      *len)
{
    cyg_devtab_entry_t *t  = (cyg_devtab_entry_t *) handle;
    dcache_info_t      *dc = (dcache_info_t *) t->priv;
    Cyg_ErrNo res = ENOERR;

#ifdef DEBUG
    diag_printf("disk cache set config\n"); 
#endif

    switch (key) {
    case CYG_IO_SET_CONFIG_DISK_CACHE_SYNC:
        res = dcache_sync(dc);
        break;

    default:
        // Pass it on to next driver level
        res = cyg_io_set_config(dc->dev_h, key, xbuf, len);
    }

    return res;
}

// --------------------------------------------------------------------
// EOF disk_cache.c
